class FileWarning(Exception):

    def __init__(self, file, message, *args, **kwargs):
        file.warn(message, *args, **kwargs)
        super().__init__(message % args)


class FileError(Exception):

    def __init__(self, file, message, *args, **kwargs):
        file.error(message, *args, **kwargs)
        super().__init__(message % args)


class FileCritical(Exception):

    def __init__(self, file, message, *args, **kwargs):
        file.critical(message, *args, **kwargs)
        super().__init__(message % args)



import sys
from os import path

import colorlog

from isimip_utils.parser import ArgumentParser

from . import VERSION
from .checks import checks
from .config import settings
from .exceptions import FileCritical, FileError, FileWarning
from .models import File, Summary
from .utils.files import walk_files
from .utils.logging import CHECKING

logger = colorlog.getLogger(__name__)


def get_parser():
    parser = ArgumentParser(prog='isimip-qc', description='Check ISIMIP files for matching protocol definitions')

    # mandatory
    parser.add_argument('schema_path', help='ISIMIP schema_path, e.g. ISIMIP3a/OutputData/water_global')

    # optional
    parser.add_argument('-c', '--copy', dest='copy', action='store_true',
                        help='copy checked files to CHECKED_PATH if no warnings or errors were found')
    parser.add_argument('-m', '--move', dest='move', action='store_true',
                        help='move checked files to CHECKED_PATH if no warnings or errors were found')
    parser.add_argument('-O', '--overwrite', dest='overwrite', action='store_true',
                        help='overwrite files in CHECKED_PATH if present. Default is False.')
    parser.add_argument('--unchecked-path', dest='unchecked_path',
                        help='base path of the unchecked files')
    parser.add_argument('--checked-path', dest='checked_path',
                        help='base path for the checked files')
    parser.add_argument('--protocol-location', dest='protocol_locations',
                        default='https://protocol.isimip.org https://protocol2.isimip.org',
                        help='URL or file path to the protocol when different from official repository')
    parser.add_argument('--log-level', dest='log_level', default='CHECKING',
                        help='log level (CRITICAL, ERROR, WARN, VRDETAIL, CHECKING, SUMMARY,'
                             ' INFO, or DEBUG) [default: CHECKING]')
    parser.add_argument('--log-path', dest='log_path',
                        help='base path for the individual log files')
    parser.add_argument('--log-path-level', dest='log_path_level', default='WARN',
                        help='log level for the individual log files [default: WARN]')
    parser.add_argument('--include', dest='include_list',
                        help='patterns of files to include. Exclude those that don\'t match any.')
    parser.add_argument('--exclude', dest='exclude_list',
                        help='patterns of files to exclude. Include only those that don\'t match any.')
    parser.add_argument('-f', '--first-file', dest='first_file', action='store_true', default=False,
                        help='only process first file found in UNCHECKED_PATH')
    parser.add_argument('-w', '--stop-on-warnings', dest='stop_warn', action='store_true', default=False,
                        help='stop execution on warnings')
    parser.add_argument('-e', '--stop-on-errors', dest='stop_err', action='store_true', default=False,
                        help='stop execution on errors')
    parser.add_argument('-r', '--minmax', dest='minmax', action='store', nargs='?', const=10, type=int,
                        help='test values for valid range (slow, argument MINMAX defaults to show the top 10 values)')
    parser.add_argument('-nt', '--skip-time-span-check', dest='time_span', action='store_true', default=False,
                        help='skip check for simulated time period')
    parser.add_argument('--fix', dest='fix', action='store_true', default=False,
                        help='try to fix warnings detected on the original files')
    parser.add_argument('--fix-datamodel', dest='fix_datamodel', action='store', nargs='?', const='nccopy', type=str,
                        help='also fix warnings on data model found using NCCOPY or CDO (slow).'
                             ' Choose preferred tool per lower case argument.')
    parser.add_argument('--check', dest='check',
                        help='perform only one particular check')
    parser.add_argument('--force-copy-move', dest='force_copy_move', action='store_true', default=False,
                        help='Copy or move files despite errors')
    parser.add_argument('-V', '--version', action='version',
                        version=VERSION)
    return parser


def init_settings(**kwargs):
    parser = get_parser()
    args = parser.get_defaults()
    args.update(kwargs)
    settings.setup(args)
    return settings


def main():
    parser = get_parser()
    args = vars(parser.parse_args())
    settings.setup(args)
    summary = Summary()

    if settings.DEFINITIONS is None:
        parser.error('no definitions could be found. Check schema_path argument.')
    if settings.PATTERN is None:
        parser.error('no pattern could be found. Check schema_path argument.')
    if settings.SCHEMA is None:
        parser.error('no schema could be found. Check schema_path argument.')

    if settings.UNCHECKED_PATH:
        if not path.exists(settings.UNCHECKED_PATH):
            logger.error('UNCHECKED_PATH does not exist:', settings.UNCHECKED_PATH)
            quit()

    if settings.CHECKED_PATH:
        if not path.exists(settings.CHECKED_PATH):
            logger.error('CHECKED_PATH does not exist:', settings.CHECKED_PATH)
            quit()

    # walk over unchecked files
    for file_path in walk_files(settings.UNCHECKED_PATH):
        if path.islink(file_path):
            continue

        logger.log(CHECKING, file_path)

        if settings.INCLUDE_LIST:
            if not any(string in str(file_path) for string in settings.INCLUDE_LIST.split(',')):
                logger.log(CHECKING, ' skipped by include option.')
                continue

        if settings.EXCLUDE_LIST:
            if any(string in str(file_path) for string in settings.EXCLUDE_LIST.split(',')):
                logger.log(CHECKING, ' skipped by exclude option.')
                continue

        if file_path.suffix in settings.PATTERN['suffix']:

            file = File(file_path)
            file.open_log()

            file.match()

            # skip opening non-NetCDF files
            if file_path.suffix not in ['.nc', '.nc4']:
                continue

            # 1st pass: perform checks
            try:
                file.open_dataset()
            except OSError:
                logger.critical('Could not open file, maybe it is corrupted, or not a NetCDF file.')
                continue

            if file.matched:

                for check in checks:
                    skip = False
                    if not settings.CHECK or check.__name__ == settings.CHECK:
                        try:
                            check(file)
                        except FileWarning:
                            pass
                        except FileError:
                            pass
                        except FileCritical:
                            skip = True
                            break

                if skip:
                    file.close_dataset()
                    continue
                else:
                    file.validate()
                    file.close_dataset()

                # log result of checks, stop if flags are set
                if file.is_clean:
                    logger.info('File has successfully passed all checks.')
                elif file.has_warnings and not file.has_errors:
                    logger.info('File passed all checks without unfixable issues.')
                elif file.has_errors:
                    logger.critical('File did not pass all checks. Unfixable issues detected.')

                if file.has_warnings and settings.STOP_WARN:
                    logger.warn('Warnings found. Exiting per -w option.')
                    sys.exit(1)

                if file.has_errors and settings.STOP_ERR:
                    logger.error('Errors found. Exiting per -e option.')
                    sys.exit(1)

                # 2nd pass: fix warnings and fixable infos
                if settings.FIX:
                    file.open_dataset(write=True)
                    if file.has_infos_fixable:
                        logger.info('Fix INFOs...')
                        file.fix_infos()
                    if file.has_warnings:
                        logger.info('Fix WARNINGs...')
                        file.fix_warnings()
                    file.close_dataset()

                # 2nd pass: fix warnings
                if file.has_warnings and settings.FIX_DATAMODEL:
                    logger.info('Fix data model...')
                    file.fix_datamodel()

                # copy/move files to checked_path
                if settings.MOVE or settings.COPY:
                    if file.is_clean or settings.FORCE_COPY_MOVE:
                        if settings.MOVE:
                            file.move()
                        elif settings.COPY:
                            file.copy()
                    else:
                        logger.warn('File has not been moved or copied due to warnings or erros found.')

                # collect stats about the file
                summary.update_specifiers(file.specifiers)
                summary.update_variables(file.specifiers)
                summary.update_experiments(file.specifiers)

            else:
                file.close_dataset()

            # close the log for this file
            file.close_log()
        else:
            logger.error('File has wrong suffix. Use "%s" for this simulation round.', settings.PATTERN['suffix'][0])

        # stop if flag is set
        if settings.FIRST_FILE:
            break

    summary.log()



import os
import shutil
from pathlib import Path

import colorlog

from ..config import settings

logger = colorlog.getLogger(__name__)


def walk_files(path):
    for root, dirs, file_names in sorted(os.walk(path)):
        for file_name in sorted(file_names):
            file_path = Path(root) / file_name
            yield file_path


def move_file(source_path, target_path, overwrite=False):
    if settings.OVERWRITE is True:
        overwrite = True

    logger.debug('source_path=%s target_path=%s', source_path, target_path)
    target_path.parent.mkdir(parents=True, exist_ok=True)
    if not target_path.is_file() or overwrite:
        logger.info('Copy file')
        shutil.move(source_path, target_path)
    else:
        logger.warn('Skip moving because target file is present and overwriting not allowed.'
                    ' Use -O to allow overwriting.')


def copy_file(source_path, target_path):
    logger.debug('source_path=%s target_path=%s', source_path, target_path)
    target_path.parent.mkdir(parents=True, exist_ok=True)
    if not target_path.is_file() or settings.OVERWRITE:
        logger.info('Copy file')
        shutil.copy(source_path, target_path)
    else:
        logger.warn('Skip copying because target file is present and overwriting not allowed.'
                    ' Use -O to allow overwriting.')



import logging
import shutil
from collections import Counter
from pathlib import Path

import colorlog
import jsonschema
from prettytable.colortable import PrettyTable

from isimip_utils.exceptions import DidNotMatch
from isimip_utils.netcdf import (
    get_dimensions,
    get_global_attributes,
    get_variables,
    open_dataset_read,
    open_dataset_write,
)
from isimip_utils.patterns import match_file

from .config import settings
from .utils.datamodel import call_cdo, call_nccopy
from .utils.experiments import get_experiment
from .utils.files import copy_file, move_file
from .utils.logging import SUMMARY


class File:

    def __init__(self, file_path):
        file_path = Path(file_path).expanduser()
        self.path = file_path.relative_to(settings.UNCHECKED_PATH)
        self.abs_path = file_path

        self.infos = []
        self.warnings = []
        self.errors = []
        self.criticals = []

        self.logger = None
        self.handler = None
        self.file_handler = None

        self.dataset = None
        self.specifiers = {}

        self.is_2d = False
        self.is_3d = False
        self.is_time_fixed = False

    @property
    def json(self):
        return {
            'dimensions': get_dimensions(self.dataset),
            'variables': get_variables(self.dataset),
            'global_attributes': get_global_attributes(self.dataset),
            'specifiers': self.specifiers
        }

    def open_log(self):
        self.logger = self.get_logger()

    def close_log(self):
        if self.handler:
            self.handler.close()

    def open_dataset(self, write=False):
        if write:
            self.dataset = open_dataset_write(self.abs_path)
        else:
            self.dataset = open_dataset_read(self.abs_path)

    def close_dataset(self):
        self.dataset.close()

    def debug(self, message, *args):
        if self.logger is not None:
            self.logger.debug(message, *args)

    def info(self, message, *args, fix=None):
        if self.logger is not None:
            self.logger.info(message, *args)
        self.infos.append((message % args, fix))

    def warn(self, message, *args, fix=None, fix_datamodel=None):
        if self.logger is not None:
            self.logger.warn(message, *args)
        self.warnings.append((message % args, fix, fix_datamodel))

    def error(self, message, *args):
        if self.logger is not None:
            self.logger.error(message, *args)
        self.errors.append(message % args)

    def critical(self, message, *args):
        if self.logger is not None:
            self.logger.critical(message, *args)
        self.criticals.append(message % args)

    def fix_infos(self):
        for info in self.infos[:]:
            message, fix = info
            if fix:
                fix['func'](*fix['args'])
                self.infos.remove(info)

    def fix_warnings(self):
        for warning in self.warnings[:]:
            message, fix, _ = warning
            if fix:
                fix['func'](*fix['args'])
                self.warnings.remove(warning)

    def fix_datamodel(self):
        # check if we need to fix using cdu
        if any(fix_datamodel for _, _, fix_datamodel in self.warnings):
            # fix using tmpfile
            tmp_abs_path = self.abs_path.parent / ('.' + self.abs_path.name + '-fix')
            if settings.FIX_DATAMODEL == 'cdo':
                if shutil.which('cdo'):
                    self.info('Rewriting file with fixed data model using "cdo"')
                    call_cdo(['--history', '-s', '-z', 'zip_5', '-f', 'nc4c', '-b', 'F32', '-k', 'grid', '-copy'],
                             self.abs_path, tmp_abs_path)
                else:
                    self.error('"cdo" is not available for execution. Please install before.')
            elif settings.FIX_DATAMODEL == 'nccopy':
                if shutil.which('nccopy'):
                    self.info('Rewriting file with fixed data model using "nccopy"')
                    call_nccopy(['-k4', '-d5'], self.abs_path, tmp_abs_path)
                else:
                    self.error('"nccopy" is not available for execution. Please install before.')
            else:
                self.error(f'"{settings.FIX_DATAMODEL}" is not a valid argument for --fix-datamodel option.'
                           ' Chose "nccopy" or "cdo"')

            if settings.FIX_DATAMODEL in ['nccopy', 'cdo']:
                # move tmp file to original file
                move_file(tmp_abs_path, self.abs_path, overwrite=True)

                # remove warnings after fix
                for warning in self.warnings[:]:
                    message, _, fix_datamodel = warning
                    if fix_datamodel:
                        self.warnings.remove(warning)

    @property
    def has_infos_fixable(self):
        for info in self.infos[:]:
            message, fix = info
            if fix:
                return bool(self.infos)

    @property
    def has_warnings(self):
        return bool(self.warnings)

    @property
    def has_errors(self):
        return bool(self.errors)

    @property
    def has_criticals(self):
        return bool(self.criticals)

    @property
    def is_clean(self):
        return not (self.has_warnings or self.has_errors or self.has_criticals)

    def get_logger(self):
        # setup a log handler for the command line and one for the file
        logger_name = str(self.path)
        logger = colorlog.getLogger(logger_name)
        logger.setLevel(settings.LOG_LEVEL)

        # do not propagate messages to the root logger,
        # which is configured in settings.setup()
        logger.propagate = False

        # add handlers
        logger.addHandler(self.get_stream_handler())
        if settings.LOG_PATH:
            self.handler = self.get_file_handler()
            logger.addHandler(self.handler)

        return logger

    def get_stream_handler(self):
        formatter = colorlog.ColoredFormatter(' %(log_color)s%(levelname)-9s: %(message)s%(reset)s')

        handler = colorlog.StreamHandler()
        handler.setLevel(settings.LOG_LEVEL)
        handler.setFormatter(formatter)

        return handler

    def get_file_handler(self):
        log_name = self.path.name.split('.')[0] + '_' + settings.NOW
        log_path = (settings.LOG_PATH / self.path.parent / log_name).with_suffix('.log')
        log_path.parent.mkdir(parents=True, exist_ok=True)

        formatter = logging.Formatter(' %(levelname)-9s: %(message)s')

        handler = logging.FileHandler(log_path, 'w')
        handler.setLevel(settings.LOG_PATH_LEVEL)
        handler.setFormatter(formatter)

        return handler

    def match(self):
        try:
            path, self.specifiers = match_file(settings.PATTERN, self.path)
            self.info('File matched naming scheme: %s.', self.specifiers)
            self.matched = True
        except DidNotMatch as e:
            self.error('File did not match naming scheme.')
            self.debug(e)
            self.matched = False

    def validate(self):
        try:
            jsonschema.validate(schema=settings.SCHEMA, instance=self.json)
        except jsonschema.exceptions.ValidationError as e:
            self.error('Failed to validate with JSON schema: %s\n%s', self.json, e)

    def copy(self):
        copy_file(self.abs_path, settings.CHECKED_PATH / self.path)

    def move(self):
        move_file(self.abs_path, settings.CHECKED_PATH / self.path)


class Summary:

    def __init__(self):
        self.specifiers = {}
        self.variables = {}
        self.experiments = Counter()

    def update_specifiers(self, specifiers):
        for identifier, specifier in specifiers.items():
            if identifier not in self.specifiers:
                self.specifiers[identifier] = Counter()
            self.specifiers[identifier][specifier] += 1

    def update_variables(self, specifiers):
        variable = specifiers.get('variable')
        if variable is not None:
            if variable not in self.variables:
                definition = settings.DEFINITIONS['variable'].get(variable)
                self.variables[variable] = {
                    'specifier': variable,
                    'sectors': definition.get('sectors'),
                    'count': 1
                }
            else:
                self.variables[variable]['count'] += 1

    def update_experiments(self, specifiers):
        experiment = get_experiment(specifiers)
        if experiment is not None:
            self.experiments[experiment] += 1

    def log_specifiers(self):
        table = PrettyTable()
        table.field_names = ['Identifier', 'Specifier', 'Count']
        table.align['Identifier'] = 'l'
        table.align['Specifier'] = 'l'
        table.align['Count'] = 'r'

        for identifier, counter in self.specifiers.items():
            for i, (specifier, count) in enumerate(counter.items()):
                table.add_row([identifier if i == 0 else '', specifier, count])

        for line in table.get_string().splitlines():
            colorlog.log(SUMMARY, line)

    def log_variables(self):
        table = PrettyTable()
        table.field_names = ['Specifier', 'Sectors', 'Count']
        table.align['Specifier'] = 'l'
        table.align['Long name'] = 'l'
        table.align['Sectors'] = 'l'
        table.align['Count'] = 'r'

        for specifier, variable in self.variables.items():
            table.add_row([specifier, ', '.join(variable.get('sectors')), variable.get('count')])

        for line in table.get_string().splitlines():
            colorlog.log(SUMMARY, line)

    def log_experiments(self):
        table = PrettyTable()
        table.field_names = ['Experiment', 'Count']
        table.align['Experiment'] = 'l'
        table.align['Count'] = 'r'

        for experiment, count in self.experiments.items():
            table.add_row([experiment, count])

        for line in table.get_string().splitlines():
            colorlog.log(SUMMARY, line)

    def log(self):
        self.log_specifiers()
        self.log_variables()
        self.log_experiments()



import math

import netCDF4
import numpy as np

from isimip_qc.config import settings
from isimip_qc.fixes import fix_set_variable_attr


def check_variable(file):
    variable = file.dataset.variables.get(file.variable_name)
    definition = settings.DEFINITIONS.get('variable', {}).get(file.specifiers.get('variable'))
    model = file.specifiers.get('model')
    climate_forcing = file.specifiers.get('climate_forcing')
    sens_scenario = file.specifiers.get('sens_scenario')

    if not variable:
        file.error('Variable %s is missing.', file.variable_name)
    elif not definition:
        file.error('Definition for variable %s is missing.', file.variable_name)
    else:
        # check file name and NetCDF variable to match each other
        if variable.name != file.variable_name:
            file.error('File name variable (%s) does not match internal variable name (%s).',
                       file.variable_name, variable.name)

        # check dtype
        if variable.dtype != 'float32':
            file.warn('%s data type is "%s" should be "float32".',
                      file.variable_name, variable.dtype, fix_datamodel=True)

        # check chunking
        chunking = variable.chunking()
        if chunking:
            # pick lat/lon sizes from dimensions definition
            if settings.SECTOR in ['marine-fishery_regional', 'water_regional', 'lakes_local', 'forestry']:
                lat_size = file.dataset.variables.get('lat').shape[0]
                lon_size = file.dataset.variables.get('lon').shape[0]
            else:
                lat_size = settings.DEFINITIONS['dimensions'].get('lat')['size']
                lon_size = settings.DEFINITIONS['dimensions'].get('lon')['size']

            # overwrite lat/lon ranges if available from climate forcing definition
            if 'grid' in settings.DEFINITIONS['climate_forcing'].get(climate_forcing):
                grid_info = settings.DEFINITIONS['climate_forcing'].get(climate_forcing)['grid']
                try:
                    lat_size = grid_info['lat_size'][sens_scenario]
                    lon_size = grid_info['lon_size'][sens_scenario]
                except KeyError:
                    lat_size = grid_info['lat_size']['default']
                    lon_size = grid_info['lon_size']['default']

            # overwrite for special cases not defined in the protocol
            if model == 'dbem':
                lat_size = 360
                lon_size = 720

            if file.is_2d:
                if chunking[0] != 1 or chunking[1] != lat_size or chunking[2] != lon_size:
                    file.warn('%s.chunking=%s should be [1, %s, %s] (with proper depencency order).',
                              file.variable_name, chunking, lat_size, lon_size, fix_datamodel=True)
                else:
                    file.info('Variable properly chunked [1, %s, %s].', lat_size, lon_size)
            if file.is_3d:
                var3d_size = file.dataset.dimensions.get(file.dim_vertical).size
                if (
                    chunking[0] != 1
                    or (chunking[1] != 1 and chunking[1] != var3d_size)
                    or chunking[2] != lat_size
                    or chunking[3] != lon_size
                ):
                    file.warn('%s.chunking=%s. Should be [1, %s, %s, %s] or [1, 1, %s, %s]'
                              ' (with proper depencency order).',
                              file.variable_name, chunking, var3d_size, lat_size, lon_size,
                              lat_size, lon_size, fix_datamodel=True)
                else:
                    file.info('Variable properly chunked [1, %s, %s, %s].', var3d_size, lat_size, lon_size)
        else:
            file.info('Variable chunking not supported by data model found.')

        # check dimensions
        definition_dimensions = tuple(definition.get('dimensions', []))

        if file.is_time_fixed:
            default_dimensions = ('lat', 'lon')
        if file.is_2d:
            default_dimensions = ('time', 'lat', 'lon')
        elif file.is_3d:
            default_dimensions = ('time', file.dim_vertical, 'lat', 'lon')

        if definition_dimensions:
            if variable.dimensions not in [definition_dimensions, default_dimensions]:
                file.error('Found %s dimensions for "%s". Must be %s.',
                           variable.dimensions, file.variable_name, default_dimensions)
        else:
            if variable.dimensions != default_dimensions:
                file.error('Found %s dimensions for "%s". Must be %s.',
                           variable.dimensions, file.variable_name, default_dimensions)

        # check standard_name
        standard_name = definition.get('standard_name')
        if standard_name:
            try:
                if variable.standard_name != standard_name:
                    file.warn(
                        'Attribute standard_name="%s" for variable "%s". Should be "%s".',
                        variable.standard_name, file.variable_name, standard_name, fix={
                            'func': fix_set_variable_attr,
                            'args': (file, file.variable_name, 'standard_name', standard_name)
                        }
                    )
            except AttributeError:
                file.warn(
                    'Attribute standard_name is missing for variable "%s". Should be "%s".',
                    file.variable_name, standard_name, fix={
                        'func': fix_set_variable_attr,
                        'args': (file, file.variable_name, 'standard_name', standard_name)
                    }
                )

        # check long_name
        long_name = definition.get('long_name')
        if long_name:
            try:
                if variable.long_name != long_name:
                    file.warn(
                        'Attribute long_name="%s" for variable "%s". Should be "%s".',
                        variable.long_name, file.variable_name, long_name, fix={
                            'func': fix_set_variable_attr,
                            'args': (file, file.variable_name, 'long_name', long_name)
                        }
                    )
            except AttributeError:
                file.warn(
                    'Attribute long_name is missing for variable "%s". Should be "%s".',
                    file.variable_name, long_name, fix={
                        'func': fix_set_variable_attr,
                        'args': (file, file.variable_name, 'long_name', long_name)
                    }
                )

        # check variable units
        units = definition.get('units')
        if units is not None:
            try:
                if variable.units != units:
                    file.error('%s.units="%s" should be "%s". Check if values are matching the unit given.',
                               file.variable_name, variable.units, units)
                else:
                    file.info('Variable unit matches protocol definition (%s).', variable.units)
            except AttributeError:
                file.warn(
                    'Variable "%s" units attribute is missing. Should be "%s".',
                    file.variable_name, units, fix={
                        'func': fix_set_variable_attr,
                        'args': (file, file.variable_name, 'units', units)
                    }
                )
        else:
            file.warn('No units information for variable "%s" in definition.', file.variable_name)

        # check _FillValue and missing_value
        for name in ['_FillValue', 'missing_value']:
            try:
                attr = variable.getncattr(name)
                if attr.dtype != variable.dtype:
                    file.error('%s data type (%s) differs from %s data type (%s).',
                               name, attr.dtype, file.variable_name, variable.dtype)
                else:
                    if not math.isclose(attr, 1e+20, rel_tol=1e-6):
                        file.error('Missing values for variable "%s": %s=%s but should be 1e+20.',
                                   file.variable_name, name, attr)
                    else:
                        file.info('Missing value attribute "%s" is properly set.', name)
            except AttributeError:
                if name == 'missing_value':
                    file.warn(
                        '"%s" attribute for variable "%s" is missing. Should be set to 1e+20f.',
                        name, file.variable_name, fix={
                            'func': fix_set_variable_attr,
                            'args': (file, file.variable_name, 'missing_value', 1e+20)
                        }
                    )
                else:
                    file.error('"%s" attribute for variable "%s" is missing. Should be set to 1e+20f and must be set'
                               ' when variable is created.', name, file.variable_name)

        # check valid range
        if settings.SECTOR == 'agriculture':
            return

        if settings.MINMAX:
            if file.is_time_fixed:
                file.warn('Valid range test for fixed data not yet implemented')
                return

            valid_min = definition.get('valid_min')
            valid_max = definition.get('valid_max')
            if (valid_min is not None) and (valid_max is not None):
                file.info('Checking values for valid minimum and maximum range defined in'
                          ' the protocol. This could take some time...')
                lat = file.dataset.variables.get('lat')
                lon = file.dataset.variables.get('lon')
                time = file.dataset.variables.get('time')

                too_low = np.argwhere(variable[:] < valid_min)
                too_high = np.argwhere(variable[:] > valid_max)

                time = file.dataset.variables.get('time')
                file.specifiers.get('time_step')

                try:
                    time_units = time.units
                except AttributeError:
                    file.warn('Can\'t check for valid ranges because of missing units attribute in time variable')
                    return

                try:
                    time_calendar = time.calendar
                except AttributeError:
                    file.warn('Can\'t check for valid ranges because of missing calendar attribute in time variable')
                    return

                if too_low.size:
                    file.warn('%i values are lower than the valid minimum (%.2E %s).',
                              too_low.shape[0], valid_min, units)
                    if settings.LOG_LEVEL == 'VRDETAIL':
                        file.warn('%i lowest values are :', min(settings.MINMAX, too_low.shape[0]))

                        too_low_list = []
                        for index in too_low[0:too_low.shape[0]]:
                            too_low_list.append([tuple(index), variable[tuple(index)].data.tolist()])
                        too_low_sorted = sorted(too_low_list, key=lambda value: value[1], reverse=False)
                        for i in range(0, min(settings.MINMAX, too_low.shape[0])):
                            if file.is_2d:
                                file.warn('date: %s, lat/lon: %4.2f/%4.2f, value: %E %s',
                                          netCDF4.num2date(time[too_low_sorted[i][0][0]], time_units, time_calendar),
                                          lat[too_low_sorted[i][0][-2]],
                                          lon[too_low_sorted[i][0][-1]],
                                          too_low_sorted[i][1], units)
                            elif file.is_3d:
                                file.warn('date: %s, lat/lon: %4.2f/%4.2f, level: %s, value: %E %s',
                                          netCDF4.num2date(time[too_low_sorted[i][0][0]], time_units, time_calendar),
                                          lat[too_low_sorted[i][0][-2]],
                                          lon[too_low_sorted[i][0][-1]],
                                          too_low_sorted[i][0][-3] + 1,
                                          too_low_sorted[i][1], units)

                if too_high.size:
                    file.warn('%i values are higher than the valid maximum (%.2E %s).',
                              too_high.shape[0], valid_max, units)
                    if settings.LOG_LEVEL == 'VRDETAIL':
                        file.warn('%i highest values are :', min(settings.MINMAX, too_high.shape[0]))

                        too_high_list = []
                        for index in too_high[0:too_high.shape[0]]:
                            too_high_list.append([tuple(index), variable[tuple(index)].data.tolist()])
                        too_high_sorted = sorted(too_high_list, key=lambda value: value[1], reverse=True)
                        for i in range(0, min(settings.MINMAX, too_high.shape[0])):
                            if file.is_2d:
                                file.warn('date: %s, lat/lon: %4.2f/%4.2f, value: %E %s',
                                          netCDF4.num2date(time[too_high_sorted[i][0][0]], time_units, time_calendar),
                                          lat[too_high_sorted[i][0][-2]],
                                          lon[too_high_sorted[i][0][-1]],
                                          too_high_sorted[i][1], units)
                            elif file.is_3d:
                                file.warn('date: %s, lat/lon: %4.2f/%4.2f, level: %s, value: %E %s',
                                          netCDF4.num2date(time[too_high_sorted[i][0][0]], time_units, time_calendar),
                                          lat[too_high_sorted[i][0][-2]],
                                          lon[too_high_sorted[i][0][-1]],
                                          too_high_sorted[i][0][-3] + 1,
                                          too_high_sorted[i][1], units)

                if not too_low.shape and not too_high.shape:
                    file.info('Values are within valid range (%.2E to %.2E).', valid_min, valid_max)

            else:
                file.warn('No min and/or max definition found for variable "%s" in protocol. Skipping test.',
                          file.variable_name)



import importlib
import inspect
import os
from pathlib import Path

checks_dir = Path(__file__).parent

# gather functions from the python files in this directory
functions = []

for root, dirs, files in os.walk(checks_dir):
    for file_name in sorted(files):
        file_path = checks_dir / root / file_name
        if file_path.suffix == '.py' and file_path.name != '__init__.py':
            parts = file_path.relative_to(checks_dir).with_suffix('').parts
            module_name = 'isimip_qc.checks.{}'.format('.'.join(parts))
            module = importlib.import_module(module_name)
            functions += list(inspect.getmembers(module, inspect.isfunction))

# find checks
checks = [function[1] for function in functions if function[0].startswith('check_')]



import subprocess


def call_cdo(args, input_file, output_file):
    args = ['cdo', *args, str(input_file), str(output_file)]
    # print(' '.join(args))
    subprocess.check_call(args)


def call_nccopy(args, input_file, output_file):
    args = ['nccopy', *args, str(input_file), str(output_file)]
    # print(' '.join(args))
    subprocess.check_call(args)



from ..fixes import fix_remove_variable_attr, fix_rename_dimension, fix_rename_variable, fix_rename_variable_attr


def check_data_model(file):
    '''
    File must use the NetCDF4 classic data model
    '''
    if file.dataset.data_model != 'NETCDF4_CLASSIC':
        file.warn('Data model is %s (not NETCDF4_CLASSIC).', file.dataset.data_model, fix_datamodel=True)
    else:
        file.info('Data model looks good (%s).', file.dataset.data_model)


def check_zip(file):
    '''
    Data variables must be compressed with at least compression level 4. Skip check for dimension variables.
    '''

    variable = file.dataset.variables.get(file.variable_name)

    try:
        zlib = variable.filters().get('zlib')
        if zlib:
            complevel = variable.filters().get('complevel')
            if complevel < 4:
                file.warn('Variable "%s" compression level is "%s". Should be >= 5.',
                          file.variable_name, complevel, fix_datamodel=True)
            else:
                file.info('Variable "%s" compression level looks good (%s)', file.variable_name, complevel)
        else:
            file.warn('Variable "%s" is not compressed.', file.variable_name, fix_datamodel=True)
    except AttributeError:
        file.warn('Variable "%s" is not compressed.', file.variable_name, fix_datamodel=True)


def check_lower_case(file):
    '''
    Internal names of dimensions and variables are lowercase.
    '''

    for dimension_name in file.dataset.dimensions:
        if not dimension_name.islower():
            file.warn('Dimension "%s" is not lower case.', dimension_name, fix={
                'func': fix_rename_dimension,
                'args': (file, dimension_name)
            })

    for variable_name, variable in file.dataset.variables.items():
        if not variable_name.islower():
            file.warn('Variable "%s" is not lower case.', variable_name, fix={
                'func': fix_rename_variable,
                'args': (file, variable_name)
            })

        for attr in variable.__dict__:
            if attr not in ['_FillValue']:
                if attr not in ['axis', 'standard_name', 'long_name', 'calendar', 'missing_value',
                                'units', 'comment', 'enteric_infection', 'description', 'unit_conversion_info',
                                'positive', 'bounds', 'classes']:
                    file.warn('Attribute "%s" for variable "%s" is not needed.', attr, variable_name, fix={
                        'func': fix_remove_variable_attr,
                        'args': (file, variable_name, attr)
                    })
                elif not attr.islower():
                    file.warn('Attribute "%s" for variable "%s" is not lower case.', attr, variable_name, fix={
                        'func': fix_rename_variable_attr,
                        'args': (file, variable_name, attr)
                    })




from isimip_qc.config import settings
from isimip_qc.fixes import fix_set_variable_attr


def check_time_variable(file):
    if file.is_time_fixed:
        return
    time = file.dataset.variables.get('time')
    time_definition = settings.DEFINITIONS['dimensions'].get('time')

    if time is None:
        file.error('Variable time is missing.')
    elif not time_definition:
        file.error('Definition for variable time is missing.')
    else:
        # check dtype
        dtypes = ['float32', 'float64']
        if time.dtype not in dtypes:
            file.warn('Data type of "time" is "%s". Should be float or double (one of %s).', time.dtype, dtypes)

        # check axis
        axis = time_definition.get('axis')
        try:
            if time.axis != axis:
                file.warn('"axis" attribute of "time" is %s. Should be "%s".', time.axis, axis, fix={
                    'func': fix_set_variable_attr,
                    'args': (file, 'time', 'axis', axis)
                })
        except AttributeError:
            file.warn('"axis" attribute of "time" is missing. Should be "%s".', axis, fix={
                'func': fix_set_variable_attr,
                'args': (file, 'time', 'axis', axis)
            })

        # check standard_name
        standard_name = time_definition.get('standard_name')
        try:
            if time.standard_name != standard_name:
                file.warn(
                    '"standard_name" attribute of "time" is "%s". Should be "%s".',
                    time.standard_name, standard_name, fix={
                        'func': fix_set_variable_attr,
                        'args': (file, 'time', 'standard_name', standard_name)
                    }
                )
        except AttributeError:
            file.warn('"standard_name" attribute of "time" is missing. Should be "%s".', standard_name, fix={
                'func': fix_set_variable_attr,
                'args': (file, 'time', 'standard_name', standard_name)
            })

        # check long_name
        long_names = time_definition.get('long_names', [])
        try:
            if time.long_name not in long_names:
                file.warn('"long_name" attribute of "time" is %s". Should be in %s.', time.long_name, long_names, fix={
                    'func': fix_set_variable_attr,
                    'args': (file, 'time', 'long_name', long_names[2])
                })
        except AttributeError:
            file.warn('"long_name" attribute of "time" is missing. Should be "%s".', long_names[2], fix={
                'func': fix_set_variable_attr,
                'args': (file, 'time', 'long_name', long_names[2])
            })

        # check units
        time_step = file.specifiers.get('time_step')
        increment = settings.DEFINITIONS['time_step'][time_step]['increment']
        minimum = settings.DEFINITIONS['time_span']['minimum']['value']

        units_templates = [
            "%s since %i-01-01",
            "%s since %i-01-01 00:00:00",
            "%s since %i-1-1",
            "%s since %i-1-1 00:00:00"
        ]

        units = [template % (increment, minimum) for template in units_templates]

        try:
            if time.units not in units:
                file.error('"units" attribute for "time" is "%s". Should be "%s".', time.units, units)
            else:
                file.info('Valid time unit found (%s)', time.units)
        except AttributeError:
            file.error('Attribute time.units is missing. Should be "%s".', units)

        # check calendars
        try:
            calendars = time_definition.get('calenders_daily', [])
            if time.calendar not in calendars:
                file.error('"calendar" attribute for "time" is "%s". Must be one of "%s".', time.calendar, calendars)
            else:
                file.info('Valid calendar found (%s)', time.calendar)
        except AttributeError:
            file.warn('"calendar" attribute for "time" is missing. Should be in "%s".', calendars)

def check_time_span_periods(file):

    if settings.TIME_SPAN:
        file.info('skipping test of covered simulation period per option')
        return

    climate_forcing = file.specifiers.get('climate_forcing')

    if 'pre-industrial' in str(file.abs_path):
        definition_startyear = settings.DEFINITIONS['time_span'].get('start_pre-ind')['value']
        definition_endyear = settings.DEFINITIONS['time_span'].get('end_pre-ind')['value']
    elif 'historical' in str(file.abs_path):
        if settings.SIMULATION_ROUND in ['ISIMIP2a', 'ISIMIP3a']:
            definition_startyear = settings.DEFINITIONS['time_span'].get('start_hist')['value'][climate_forcing]
            definition_endyear = settings.DEFINITIONS['time_span'].get('end_hist')['value'][climate_forcing]
        elif settings.SIMULATION_ROUND in ['ISIMIP2b', 'ISIMIP3b']:
            definition_startyear = settings.DEFINITIONS['time_span'].get('start_hist')['value']
            definition_endyear = settings.DEFINITIONS['time_span'].get('end_hist')['value']
    elif 'future' in str(file.abs_path):
        definition_startyear = settings.DEFINITIONS['time_span'].get('start_fut')['value']
        definition_endyear = settings.DEFINITIONS['time_span'].get('end_fut')['value']
    else:
        file.warn('Skipping check for simulation period as the period itself could not be'
                  ' determined from the file path (pre-industrial, historical or future).')
        return

    file_startyear = file.specifiers.get('start_year')
    file_endyear = file.specifiers.get('end_year')

    time_resolution = file.specifiers.get('time_step')

    if time_resolution not in ['daily', 'fixed']:
        if definition_startyear != file_startyear or definition_endyear != file_endyear:
            file.warn('time period covered by file (%s-%s) does not match input data time span (%s-%s).'
                      ' Ensure to prepare the full period for all variables using the latest input data set.',
                      file_startyear, file_endyear, definition_startyear, definition_endyear)
        else:
            file.info('File is covering the full simulation period (by file name)')
    else:
        if 'historical' in str(file.abs_path):
            if settings.SIMULATION_ROUND == 'ISIMIP3a' and climate_forcing == '20crv3-era5':
                last_file_startyear = 2021
            else:
                last_file_startyear = 2011

            if file_startyear == last_file_startyear:
                if definition_endyear != file_endyear:
                    file.warn('Last year of time period covered by file (%s) does not match end of input'
                              ' data time span (%s). Ensure to prepare the full period for all variables'
                              ' using the latest input data set.',
                              file_endyear, definition_endyear)



from ..exceptions import FileCritical
from ..utils.experiments import get_experiment


def check_experiment(file):
    if file.specifiers.get("time_step") == "fixed":
        return

    experiment = get_experiment(file.specifiers)
    if experiment is None:
        raise FileCritical(file, "No valid experiment found for this sector and {climate_scenario}, "\
                           "{soc_scenario}, {sens_scenario} between {start_year} and {end_year}. skipping..."\
                           .format(**file.specifiers)
        )

    else:
        file.info("Experiment looks good (%s).", experiment)



from datetime import datetime
from pathlib import Path

import colorlog

from isimip_utils.config import Settings as BaseSettings
from isimip_utils.decorators import cached_property
from isimip_utils.fetch import fetch_definitions, fetch_pattern, fetch_schema

logger = colorlog.getLogger(__name__)


class Settings(BaseSettings):

    def setup(self, args):
        super().setup(args)

        self.NOW = datetime.utcnow().strftime("%Y%m%dT%H%MZ")

        # set create pathes and set default values
        if self.UNCHECKED_PATH is not None:
            self.UNCHECKED_PATH = Path(self.UNCHECKED_PATH).expanduser()
        else:
            self.UNCHECKED_PATH = Path().cwd()

        if self.CHECKED_PATH is not None:
            self.CHECKED_PATH = Path(self.CHECKED_PATH).expanduser()

        self.LOG_LEVEL = self.LOG_LEVEL.upper()
        self.LOG_PATH_LEVEL = self.LOG_PATH_LEVEL.upper()
        if self.LOG_PATH is not None:
            self.LOG_PATH = Path(self.LOG_PATH).expanduser()

        # setup logs
        colorlog.basicConfig(level=self.LOG_LEVEL,
                             format=' %(log_color)s%(levelname)-8s : %(message)s%(reset)s')

        # set the simulation_round, product and sector
        self.SIMULATION_ROUND, self.PRODUCT, self.SECTOR = Path(self.SCHEMA_PATH).parts[0:3]

        # log settings
        colorlog.debug(self)

    @cached_property
    def DEFINITIONS(self):
        assert self.PROTOCOL_LOCATIONS is not None, 'PROTOCOL_LOCATIONS is not set'
        return fetch_definitions(self.PROTOCOL_LOCATIONS.split(), self.SCHEMA_PATH)

    @cached_property
    def PATTERN(self):
        assert self.PROTOCOL_LOCATIONS is not None, 'PROTOCOL_LOCATIONS is not set'
        return fetch_pattern(self.PROTOCOL_LOCATIONS.split(), self.SCHEMA_PATH)

    @cached_property
    def SCHEMA(self):
        assert self.PROTOCOL_LOCATIONS is not None, 'PROTOCOL_LOCATIONS is not set'
        return fetch_schema(self.PROTOCOL_LOCATIONS.split(), self.SCHEMA_PATH)


settings = Settings()



from ..config import settings


def get_experiment(specifiers):
    climate_scenario = specifiers.get('climate_scenario')
    soc_scenario = specifiers.get('soc_scenario')
    sens_scenario = specifiers.get('sens_scenario')

    for experiment_specifier, experiment_values in settings.DEFINITIONS.get('experiments').items():
        for period_specifier, period_values in settings.DEFINITIONS.get('period', []).items():
            if (period_values.get('start_year') <= specifiers.get('start_year') and
                    period_values.get('end_year') >= specifiers.get('end_year')):

                experiment_period = experiment_values.get(period_specifier, {})

                if (isinstance(experiment_period, dict) and
                        experiment_period.get('climate') == climate_scenario and
                        experiment_period.get('soc') == soc_scenario and
                        (sens_scenario == 'default' or
                            experiment_period.get('climate_sens') == sens_scenario or
                            experiment_period.get('soc_sens') == sens_scenario)):
                    return experiment_specifier



from ..config import settings


def check_lon_dimension(file):
    model = file.specifiers.get('model')
    climate_forcing = file.specifiers.get('climate_forcing')
    sens_scenario = file.specifiers.get('sens_scenario')

    if settings.SECTOR not in ['marine-fishery_regional', 'water_regional', 'lakes_local', 'forestry']:

        if file.dataset.dimensions.get('lon') is None:
            file.error('Longitude dimension "lon" is missing.')
        else:
            if model == 'dbem':
                lon_size = 720
            else:
                lon_size = settings.DEFINITIONS['dimensions'].get('lon')['size']

            # pick longitude size if available from data set definition
            if 'grid' in settings.DEFINITIONS['climate_forcing'].get(climate_forcing):
                grid_info = settings.DEFINITIONS['climate_forcing'].get(climate_forcing)['grid']['lon_size']
                try:
                    lon_size = grid_info[sens_scenario]
                except KeyError:
                    lon_size = grid_info['default']

            if lon_size != file.dataset.dimensions.get('lon').size:
                file.warn('Unexpected number of longitudes found (%s). Should be %s',
                          file.dataset.dimensions.get('lon').size, lon_size)
            else:
                file.info('%s longitudes defined.', lon_size)


def check_lat_dimension(file):
    model = file.specifiers.get('model')
    climate_forcing = file.specifiers.get('climate_forcing')
    sens_scenario = file.specifiers.get('sens_scenario')

    if settings.SECTOR not in ['marine-fishery_regional', 'water_regional', 'lakes_local', 'forestry']:

        if file.dataset.dimensions.get('lat') is None:
            file.error('Latitude dimension "lat" is missing.')
        else:
            if model == 'dbem':
                lat_size = 360
            else:
                lat_size = settings.DEFINITIONS['dimensions'].get('lat')['size']

            # pick latitude size if available from data set definition
            if 'grid' in settings.DEFINITIONS['climate_forcing'].get(climate_forcing):
                grid_info = settings.DEFINITIONS['climate_forcing'].get(climate_forcing)['grid']['lat_size']
                try:
                    lat_size = grid_info[sens_scenario]
                except KeyError:
                    lat_size = grid_info['default']

            if lat_size != file.dataset.dimensions.get('lat').size:
                file.warn('Unexpected number of latitudes found (%s). Should be %s',
                          file.dataset.dimensions.get('lat').size, lat_size)
            else:
                file.info('%s latitudes defined.', lat_size)


def check_time_dimension(file):
    if not file.is_time_fixed and file.dataset.dimensions.get('time') is None:
        file.error('Dimension "time" is missing.')


def check_depth_dimension(file):
    if file.is_3d:
        if file.dataset.dimensions.get(file.dim_vertical) is None:
            file.error('Valid 4th dimension is missing. Should be of of [depth, bins]. Found "%s" instead.',
                       file.dim_vertical)


def check_dimensions(file):
    # check dimension order
    variable = file.dataset.variables.get(file.variable_name)

    dim_len = len(variable.dimensions)
    if file.is_time_fixed:
        if variable.dimensions[0] != 'lat' or variable.dimensions[1] != 'lon':
            file.error('Dimension order for variable "%s" is %s. Should be ["lat", "lon"].',
                       file.variable_name, variable.dimensions)
        else:
            file.info('Dimensions for variable "%s" look good: %s.',
                      file.variable_name, variable.dimensions)
    elif file.is_2d:
        if variable.dimensions[0] != 'time' or variable.dimensions[1] != 'lat' or variable.dimensions[2] != 'lon':
            file.error('Dimension order for variable "%s" is %s. Should be ["time", "lat", "lon"].',
                       file.variable_name, variable.dimensions)
        else:
            file.info('Dimensions for variable "%s" look good: %s.',
                      file.variable_name, variable.dimensions)
    elif file.is_3d:
        if variable.dimensions[0] != 'time' or variable.dimensions[2] != 'lat' or variable.dimensions[3] != 'lon':
            file.error('Dimension order for variable "%s" is %s. Should be ["time", "%s" , "lat", "lon"].',
                       file.variable_name, variable.dimensions, file.dim_vertical)
        else:
            file.info('Dimensions for variable "%s" look good.', file.variable_name)
    else:
        file.error('Variable "%s" neither holds 2d or 3d data. (dim=%s)', dim_len)

    for dimension_name, dimension in file.dataset.dimensions.items():
        dimension_definition = settings.DEFINITIONS['dimensions'].get(dimension_name)

        if dimension_definition:
            # size of lat and lon are checked above
            if dimension_definition.get('specifier') not in ['lat', 'lon']:
                size = dimension_definition.get('size')
                if size and dimension.size != size:
                    file.error('Size of "%s" dimension is %s. Must be %s.', dimension_name, dimension.size, size)
        else:
            file.error('"%s" is not a valid dimension name as per protocol.', dimension_name)



from isimip_qc.config import settings
from isimip_qc.fixes import fix_set_variable_attr


def check_3d_variable(file):

    def check_attribute(var3d, attr_type, attribute):
        try:
            var3d_attr = getattr(var3d, attr_type)
            if var3d_attr != attribute:
                file.warn('Attribute %s.%s="%s". Should be "%s".', var3d.name, attr_type, var3d_attr, attribute, fix={
                    'func': fix_set_variable_attr,
                    'args': (file, var3d.name, attr_type, attribute)
                })
        except AttributeError:
            file.warn('Attribute %s.%s is missing. Should be "%s".', var3d.name, attr_type, attribute, fix={
                'func': fix_set_variable_attr,
                'args': (file, var3d.name, attr_type, attribute)
            })

    if file.is_3d:

        var3d = file.dataset.variables.get(file.dim_vertical)
        var3d_definition = settings.DEFINITIONS['dimensions'].get(file.dim_vertical)

        # check if vertical dimension ha a variable associated
        if var3d is None and file.dim_vertical:
            file.error('Found dimension (%s) but no variable linked to it. Introduce "%s" as variable also.',
                       file.dim_vertical, file.dim_vertical)

        # check for definition in protocol for 3d variable
        if not var3d_definition:
            file.error('No definition for variable "%s" in protocol.', file.dim_vertical)

        # do checks of definition and vertical variable were found
        if var3d and var3d_definition:
            # check dtype
            dtypes = ['float32', 'float64']
            if var3d.dtype not in dtypes:
                file.warn('%s.datatype="%s" should be in %s.', file.dim_vertical, var3d.dtype, dtypes)

            # check attributes
            for attribute in ['axis', 'standard_name', 'long_name', 'units']:
                if settings.SIMULATION_ROUND in ['ISIMIP2a', 'ISIMIP2b'] and var3d.name == 'levlak':
                    continue
                attr_definition = var3d_definition.get(attribute)
                check_attribute(var3d, attribute, attr_definition)

            if file.dim_vertical == 'depth':
                # check direction of depth dimension
                depth_first = file.dataset.variables.get(file.dim_vertical)[0]
                depth_last = file.dataset.variables.get(file.dim_vertical)[-1]

                if depth_first > depth_last:
                    file.warn('Depths in wrong order. Should increase with depth . (found %s to %s)',
                              depth_first, depth_last)
                else:
                    file.info('Depths order looks good (positive down).')

            # for lakes sector
            if file.dim_vertical == 'levlak':
                if len(file.dataset.variables.get(file.dim_vertical).shape) == 1:
                    levlak_first = file.dataset.variables.get(file.dim_vertical)[0]
                    levlak_last = file.dataset.variables.get(file.dim_vertical)[-1]

                    if levlak_first > levlak_last:
                        file.warn('"levlak" in wrong order. Should increase with depth . (found %s to %s)',
                                  levlak_first, levlak_last)
                    else:
                        file.info('"levlak" order looks good (positive down).')

                if settings.SIMULATION_ROUND not in ['ISIMIP2a', 'ISIMIP2b']:

                    depth_file = file.dataset.variables.get('depth')

                    if depth_file is None:
                        file.warn('Variable "depth" not found. Introduce layer vertical center'
                                  ' depths in [m] as depth(levlak,lat,lon) or depth(time,levlak,lat,lon)')
                    else:
                        if len(depth_file.dimensions) == 4:
                            if depth_file.dimensions[1] != 'levlak':
                                file.error('Time varying "depth" variable has no dependency for "levlak"'
                                           ' level index. Expecting: depth(time,levlak,lat,lon)')
                        elif len(depth_file.dimensions) == 3:
                            if depth_file.dimensions[0] != 'levlak':
                                file.error('Fixed-time "depth" variable has no dependency for "levlak"'
                                           ' level index. Expecting: depth(levlak,lat,lon)')
                        elif len(depth_file.dimensions) == 1:
                            if depth_file.dimensions[0] != 'levlak':
                                file.error('Globally fixed "depth" variable has no dependency for "levlak"'
                                           ' level index. Expecting: depth(levlak)')
                        else:
                            file.error('No proper "levlak" dependency found in "depth" variable.')

                        depth_definition = settings.DEFINITIONS['dimensions'].get('depth')

                        if depth_definition is None:
                            file.warn('Dimension "depth" is not yet defined in protocol. Skipping'
                                      ' attribute checks for "depth".')
                        else:
                            for attribute in ['axis', 'standard_name', 'long_name', 'units']:
                                attr_definition = depth_definition.get(attribute)
                                check_attribute(depth_file, attribute, attr_definition)



import calendar

import netCDF4

from isimip_qc.config import settings


def check_time_resolution(file):
    if settings.SECTOR == 'agriculture':
        return

    if file.is_time_fixed:
        return

    time = file.dataset.variables.get('time')
    time_definition = settings.DEFINITIONS['dimensions'].get('time')
    time_resolution = file.specifiers.get('time_step')

    try:
        time_units = time.units
    except AttributeError:
        file.warn('Can\'t check for number of time steps because of missing time.units attribute')
        return

    try:
        time_calendar = time.calendar
    except AttributeError:
        file.warn('Can\'t check for number of time steps because of missing time.calendar attribute')
        return

    if time_resolution in ['monthly', 'annual'] and time_calendar == '360_day':
        file.error('360_day calendar is not allowed for monthly and annual data anymore.'
                   ' Use one of ["standard", "proleptic_gregorian", "365_day", "366_day"]')
        return

    if 'days' not in time_units:
        file.error('time index has to be in daily increments for all calendars used'
                   ' (time.unit to be like "days since ...".')
        return

    if file.dataset.data_model in ['NETCDF4', 'NETCDF4_CLASSIC']:

        if all([time, time_definition, time_resolution, time_units, time_calendar]):
            # first and last year from file name specifiers must match those from internal time axis
            # number of time steps must match those expected from the time axis
            time_steps = time.shape[0]

            if time_resolution in ['daily', 'monthly', 'annual']:
                if settings.SECTOR == 'agriculture' and time_resolution == 'annual':
                    ref_year = int(time.units.split()[3].split("-")[0])
                    startyear_nc = ref_year + int(time[0])
                    endyear_nc = ref_year + int(time[-1])
                else:
                    firstdate_nc = netCDF4.num2date(time[0], time_units, time_calendar)
                    lastdate_nc = netCDF4.num2date(time[time_steps-1], time_units, time_calendar)
                    startyear_nc = firstdate_nc.year
                    endyear_nc = lastdate_nc.year

            startyear_file = int(file.specifiers.get('start_year'))
            endyear_file = int(file.specifiers.get('end_year'))
            nyears_file = endyear_file - startyear_file + 1

            if startyear_nc != startyear_file or endyear_nc != endyear_file:
                file.error('Start and/or end year of NetCDF time axis (%s-%s) doesn\'t'
                           ' match period defined in file name (%s-%s)',
                           startyear_nc, endyear_nc, startyear_file, endyear_file)
            else:
                file.info('Time period covered by this file matches the internal time axis (%s-%s)',
                          startyear_nc, endyear_nc)

            if time_resolution == 'daily':
                if time_calendar in ['proleptic_gregorian', 'standard']:
                    time_days = 0
                    for year in range(startyear_file, endyear_file+1):
                        if calendar.isleap(year):
                            time_days += 366
                        else:
                            time_days += 365
                elif time_calendar == '366_day':
                    time_days = nyears_file * 366
                elif time_calendar in ['365_day', 'noleap']:
                    time_days = nyears_file * 365
                elif time_calendar == '360_day':
                    time_days = nyears_file * 360

                if time_days != time_steps:
                    file.error('Number of internal time steps (%s) does not match the expected'
                               ' number from the file name specifiers (%s). ("%s" calendar found)',
                               time_steps, time_days, time_calendar)
                else:
                    file.info('Correct number of time steps (%s) given the defined calendar (%s)',
                              time_steps, time_calendar)
            elif time_resolution == 'monthly':
                time_months = nyears_file * 12
                if time_months != time_steps:
                    file.error('Number of internal time steps (%s) does not match the expected'
                               ' number from the file name specifiers (%s).', time_steps, time_months)
                else:
                    file.info('Correct number of time steps (%s).', time_steps)
            elif time_resolution == 'annual':
                if nyears_file != time_steps:
                    file.error('Number of internal time steps (%s) does not match the expected'
                               ' number from the file name specifiers (%s).', time_steps, nyears_file)
                else:
                    file.info('Correct number of time steps (%s).', time_steps)
    else:
        file.warn('Could not check for the correct number of time steps because of wrong'
                  ' data model (%s). Has to be NETCDF4_CLASSIC.', file.dataset.data_model)



import logging as python_logging

VRDETAIL = 23
CHECKING = 22
SUMMARY = 21

python_logging.addLevelName(VRDETAIL, 'VRDETAIL')
python_logging.addLevelName(CHECKING, 'CHECKING')
python_logging.addLevelName(SUMMARY, 'SUMMARY')



import uuid
from datetime import datetime
from email.utils import parseaddr

from .. import __version__
from ..config import settings
from ..fixes import fix_remove_global_attr, fix_set_global_attr


def check_isimip_id(file):
    try:
        isimip_id = file.dataset.getncattr('isimip_id')
        file.info('Global attribute "isimip_id" found (%s).', isimip_id)
    except AttributeError:
        file.info('Global attribute "isimip_id" not yet set.', fix={
            'func': fix_set_global_attr,
            'args': (file, 'isimip_id', str(uuid.uuid4()))
        })


def check_isimip_qc_version(file):
    try:
        version = file.dataset.getncattr('isimip_qc_version')
        if version == __version__:
            file.info('Global attribute "isimip_qc_version" matches current tool version (%s).',
                      version)
        else:
            file.info('Global attribute "isimip_qc_version" (%s) does not match tool current tool version (%s).',
                      version, __version__, fix={
                          'func': fix_set_global_attr,
                          'args': (file, 'isimip_qc_version', __version__)
                      })
    except AttributeError:
        file.info('Global attribute "isimip_qc_version" not yet set.',
                  fix={
                      'func': fix_set_global_attr,
                      'args': (file, 'isimip_qc_version', __version__)
                  })


def check_isimip_protocol_version(file):
    protocol_version = settings.DEFINITIONS['commit']

    try:
        version = file.dataset.getncattr('isimip_protocol_version')
        if version == protocol_version:
            file.info('Global attribute "isimip_protocol_version" matches current protocol version (%s).',
                      version)
        else:
            file.info(
                'Global attribute "isimip_protocol_version" (%s) does not match tool current protocol version (%s).',
                version, protocol_version, fix={
                    'func': fix_set_global_attr,
                    'args': (file, 'isimip_protocol_version', protocol_version)
                }
            )
    except AttributeError:
        file.info('Global attribute "isimip_protocol_version" not yet set.',
                  fix={
                      'func': fix_set_global_attr,
                      'args': (file, 'isimip_protocol_version', protocol_version)
                  })


def check_institution(file):
    try:
        file.dataset.getncattr('institution')
    except AttributeError:
        file.error('Global attribute "institution" is missing.')


def check_contact(file):
    try:
        contact = file.dataset.getncattr('contact')
        name, address = parseaddr(contact)
        if not address:
            file.error('Global attribute "contact" does not contain a proper address (%s).', contact)
        else:
            file.info('Global attribute "contact" looks good.')
    except AttributeError:
        file.error('Global attribute "contact" is missing.')


def check_isimip_qc_date(file):
    datetime_now = datetime.utcnow().strftime("%Y-%m-%d %H:%M UTC")
    try:
        date = file.dataset.getncattr('isimip_qc_pass_date')
        if date is not None:
            file.info('Global attribute "isimip_qc_pass_date" is set to "%s".',
                      date, fix={
                          'func': fix_set_global_attr,
                          'args': (file, 'isimip_qc_pass_date', datetime_now)
                      })
    except AttributeError:
        file.info('Global attribute "isimip_qc_pass_date" not yet set.',
                  fix={
                      'func': fix_set_global_attr,
                      'args': (file, 'isimip_qc_pass_date', datetime_now)
                  })


def check_history(file):
    try:
        file.dataset.getncattr('history')
        file.warn('Global attribute "history" is set and will get removed.',
                  fix={
                      'func': fix_remove_global_attr,
                      'args': (file, 'history')
                  })
    except AttributeError:
        pass



def fix_rename_dimension(file, dimension_name, new_dimension_name):
    file.info('Renaming dimension "%s" -> "%s".', dimension_name, new_dimension_name)
    file.dataset.dimensions[dimension_name].renameDimension(dimension_name, new_dimension_name)


def fix_rename_variable(file, variable_name, new_variable_name):
    file.info('Renaming variable "%s" -> "%s".', variable_name, new_variable_name)
    file.dataset.variables[variable_name].renameVariable(variable_name, new_variable_name)


def fix_set_variable_attr(file, variable_name, attr_name, value):
    file.info('Setting attribute "%s.%s=%s"', variable_name, attr_name, value)
    file.dataset.variables[variable_name].setncattr(attr_name, value)


def fix_rename_variable_attr(file, variable_name, attr_name, new_attr_name):
    file.info('Renaming attribute "%s.%s" -> "%s.%s".', variable_name, attr_name, variable_name, new_attr_name)
    file.dataset.variables[variable_name].renameAttribute(attr_name, new_attr_name)


def fix_remove_variable_attr(file, variable_name, attr_name):
    file.info('Removing attribute "%s.%s"', variable_name, attr_name)
    file.dataset.variables[variable_name].delncattr(attr_name)


def fix_set_global_attr(file, attr_name, value):
    file.info('Setting global attribute "%s=%s"', attr_name, value)
    file.dataset.setncattr(attr_name, value)


def fix_rename_global_attr(file, attr_name, new_attr_name):
    file.info('Renaming global attribute "%s" -> "%s".', attr_name, new_attr_name)
    file.dataset.renameAttribute(attr_name, new_attr_name)


def fix_remove_global_attr(file, attr_name):
    file.info('Removing global attribute "%s"', attr_name)
    file.dataset.delncattr(attr_name)



import numpy as np

from isimip_qc.config import settings
from isimip_qc.fixes import fix_set_variable_attr


def check_latlon_variable(file):
    model = file.specifiers.get('model')
    climate_forcing = file.specifiers.get('climate_forcing')
    sens_scenario = file.specifiers.get('sens_scenario')

    for variable in  ['lat', 'lon']:
        var = file.dataset.variables.get(variable)
        var_definition = settings.DEFINITIONS['dimensions'].get(variable)

        if var is None:
            file.error('Variable "%s" is missing.', variable)
        elif not var_definition:
            file.error('Definition for variable "s%" is missing.', variable)
        else:
            # check dtype
            dtypes = ['float32', 'float64']
            if var.dtype not in dtypes:
                file.warn('Data type of "%s" is "%s". Should be float or double (one of %s).',
                          variable, var.dtype, dtypes)

            # check axis
            axis = var_definition.get('axis')
            try:
                if var.axis != axis:
                    file.warn('"axis" attribute of "%s" is %s. Should be "%s".', variable, var.axis, axis, fix={
                        'func': fix_set_variable_attr,
                        'args': (file, variable, 'axis', axis)
                    })
            except AttributeError:
                file.warn('"axis" attribute of "%s" is missing. Should be "%s".', variable, axis, fix={
                    'func': fix_set_variable_attr,
                    'args': (file, variable, 'axis', axis)
                })

            # check standard_name
            standard_name = var_definition.get('standard_name')
            try:
                if var.standard_name != standard_name:
                    file.warn(
                        '"standard_name" attribute of "%s" is "%s". Should be "%s".',
                        variable, var.standard_name, standard_name, fix={
                            'func': fix_set_variable_attr,
                            'args': (file, variable, 'standard_name', standard_name)
                        }
                    )
            except AttributeError:
                file.warn(
                    '"standard_name" attribute of "%s" is missing. Should be "%s".',
                    variable, standard_name, fix={
                        'func': fix_set_variable_attr,
                        'args': (file, variable, 'standard_name', standard_name)
                    }
                )

            # check long_name
            long_names = var_definition.get('long_names', [])
            try:
                if var.long_name not in long_names:
                    file.warn(
                        '"long_name" attribute of "%s" is %s". Should be in %s.',
                        variable, var.long_name, long_names, fix={
                            'func': fix_set_variable_attr,
                            'args': (file, variable, 'long_name', long_names[0])
                        }
                    )
            except AttributeError:
                file.warn(
                    '"long_name" attribute of "%s" is missing. Should be "%s".',
                    variable, long_names[0], fix={
                        'func': fix_set_variable_attr,
                        'args': (file, variable, 'long_name', long_names[0])
                    }
                )

            # check units
            units = var_definition.get('units')
            try:
                if var.units != units:
                    file.warn(
                        '"units" attribute for "%s" is "%s". Should be "%s".',
                        variable, var.units, units, fix={
                            'func': fix_set_variable_attr,
                            'args': (file, variable, 'units', units)
                        }
                    )
            except AttributeError:
                file.warn(
                    '"units" attribute for "%s" is missing. Should be "%s".',
                    variable, units, fix={
                        'func': fix_set_variable_attr,
                        'args': (file, variable, 'units', units)
                    }
                )

            if settings.SECTOR not in ['marine-fishery_regional', 'water_regional', 'lakes_local', 'forestry']:

                # pick lat/lon ranges if available from dimensions definition
                minimum = var_definition.get('minimum')
                maximum = var_definition.get('maximum')

                # overwrite lat/lon ranges if available from climate forcing definition
                if 'grid' in settings.DEFINITIONS['climate_forcing'].get(climate_forcing):
                    grid_info = settings.DEFINITIONS['climate_forcing'].get(climate_forcing)['grid']
                    try:
                        if var.name == 'lat':
                            minimum = grid_info['lat_min'][sens_scenario]
                            maximum = grid_info['lat_max'][sens_scenario]
                        elif var.name == 'lon':
                            minimum = grid_info['lon_min'][sens_scenario]
                            maximum = grid_info['lon_max'][sens_scenario]
                    except KeyError:
                        if var.name == 'lat':
                            minimum = grid_info['lat_min']['default']
                            maximum = grid_info['lat_max']['default']
                        elif var.name == 'lon':
                            minimum = grid_info['lon_min']['default']
                            maximum = grid_info['lon_max']['default']

                # overwrite for special cases not defined in the protocol
                if model == 'dbem':
                    if var.name == 'lat':
                        minimum = -89.75
                        maximum = 89.75
                    elif var.name == 'lon':
                        minimum = -179.75
                        maximum = 179.75
                elif model == 'dbpm' or model == 'zoomss':
                    if climate_forcing == 'ipsl-cm6a-lr' and var.name == 'lon':
                        minimum = -180.
                        maximum = 179.

                if np.min(var) != minimum:
                    file.error('First value of variable "%s" is %s. Must be %s.', variable, np.min(var), minimum)

                if np.max(var) != maximum:
                    file.error('Last value of variable "%s" is %s. Must be %s.', variable, np.max(var), maximum)

                # check latitude order
                if variable == 'lat':
                    lat_first = file.dataset.variables.get('lat')[0]
                    lat_last = file.dataset.variables.get('lat')[-1]
                    if lat_first < lat_last:
                        file.warn('Latitudes in wrong order. Index should range from north to south. (found %s to %s)',
                                  lat_first, lat_last)
                    else:
                        file.info('Latitude index order looks good (N to S).')



from isimip_qc.config import settings

from ..exceptions import FileCritical, FileWarning


def check_3d(file):
    crop = file.specifiers.get('crop')
    irrigation = file.specifiers.get('irrigation')
    pft = file.specifiers.get('pft')
    species = file.specifiers.get('species')
    pool = file.specifiers.get('pool')
    pt = file.specifiers.get('pt')

    file.variable_name = file.specifiers.get('variable')

    if crop:
        file.variable_name = file.variable_name + '-' + file.specifiers.get('crop')
    if irrigation:
        file.variable_name = file.variable_name + '-' + file.specifiers.get('irrigation')
    if species:
        file.variable_name = file.variable_name + '-' + file.specifiers.get('species')
    if pt:
        file.variable_name = file.variable_name + '-' + file.specifiers.get('pt')
    if pool:
        file.variable_name = file.variable_name + '-' + file.specifiers.get('pool')
    if pft:
        file.variable_name = file.variable_name + '-' + file.specifiers.get('pft')

    try:
        variable = file.dataset.variables.get(file.variable_name)
        if variable is None:
            raise FileCritical(file,
                               'Variable "%s" from file name not found inside the file!'
                               ' Check NetCDF header. Stopping tool!',
                               file.variable_name
            )
            raise SystemExit(1)
    except AttributeError as e:
        raise FileCritical(file,
                           'Variable "%s" from file name not found inside the file!'
                           ' Check NetCDF header. Stopping tool!',
                           file.variable_name
        ) from e
        raise SystemExit(1) from e

    definition = settings.DEFINITIONS.get('variable', {}).get(file.specifiers.get('variable'))
    if definition is None:
        raise FileCritical(file,
                           'Variable %s not defined for sector %s. skipping...',
                           file.variable_name, settings.SECTOR
        )

    # check for number of variable dependencies

    file_dim_len = len(variable.dimensions)

    if 'dimensions' in definition:
        settings_dim_len = len(definition['dimensions'])
        dims_expected = str(definition['dimensions']).replace('\'', '')
    else:
        settings_dim_len = 3
        dims_expected = '[time, lat, lon]'

    if file_dim_len != settings_dim_len:
        raise FileCritical(file,
                           'Dimension missing (%s found, %s expected). Declare variable as %s%s.',
                           file_dim_len, settings_dim_len, file.variable_name, dims_expected
        )

    # detect 2d or 3d data
    file.is_time_fixed = False

    if file.specifiers.get('time_step') == 'fixed':
        file.is_time_fixed = True
        if file_dim_len > 2:
            raise FileCritical(file,
                               'File has fixed data but more than 2 dimensions. Remove "time" dimension if present.'
            )
            return
    elif file_dim_len == 3:
        file.is_2d = True
    elif file_dim_len == 4:
        file.is_3d = True
        if variable.dimensions[1] in ['time', 'lat', 'lon']:
            for pos in range(0, file_dim_len):
                if variable.dimensions[pos] not in ['time', 'lat', 'lon']:
                    break
            raise FileCritical(file,
                               'Variable "%s" has "time", "lat" or "lon" as second dependeny.'
                               ' Depencency order must be [time, %s, lat, lon]. "time" is first always.',
                               file.variable_name, variable.dimensions[pos]
            )
            file.dim_vertical = variable.dimensions[pos]
        else:
            file.dim_vertical = variable.dimensions[1]

    # check if vertical bounds were defined
    if file.is_3d:
        if variable.dimensions[1] in ['depth', 'levlak']:
            if definition.get('bounds') is None:
                raise FileWarning(file,
                                  'No vertical boundaries defined for %s dimension. ' +
                                  'Consider adding depth_bnds(depth, bnds). ' +
                                  'See examples at https://bit.ly/ncdf-bounds', variable.dimensions[1]
                )
